from types import FunctionType
import pennylane as qml
import numpy as np
import random
import sys

sys.path += ['.', './ciruits/']

from hqm.circuits.circuit import QuantumCircuit

class RandomCircuitNoiseGenerator(QuantumCircuit):
    '''
        This class implements a quantum noise generator using a random circuit for each sample.
    '''
    
    def __init__(self, location : float, scale : float, n_qubits: int, n_layers : int, dev : qml.devices = None) -> None:
        '''
            RandomCircuitNoiseGenerator constructor.  

            Parameters:  
            -----------
            - location : float  
                location parameter translate the pdf, relative to the standard normal distribution
            - scale : float  
                scale parameter that stretches the pdf. The greater the magnitude, the greater the stretching. 
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'. Recommendent qml.device("default.mixed", wires=1, shots=1000).
            
            Returns:  
            --------  
            Nothing, a RandomCircuitNoiseGenerator object will be created.  
        '''

        super().__init__(n_qubits=n_qubits, n_layers=n_layers, dev=dev)

        self.shots     = dev.shots
        self.location  = location 
        self.scale     = scale
        self.n_qubits  = n_qubits
        self.n_layers  = n_layers
        self.dev       = dev
        
    def generate_noise(self):
        '''
            GaussianLikeNoiseGenerator that generates on observable  

            Parameters:  
            -----------
            Nothing
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise observable
        '''
        expval = self.circ(self.dev)
        n = self.location + self.scale*expval
        return n

    
    def generate_noise_array(self, shape:tuple):
        '''
            GaussianLikeNoiseGenerator that generates a matrix of observables

            Parameters:  
            -----------
            shape : tuple
                shape of the desired maxtrix of observables
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise matrix of observables
        '''
        n = []
        for _ in range(np.prod(shape)): n.append(self.generate_noise())
        n = np.array(n)
    
        return n.reshape(shape)
    
    def circ(self, dev : qml.devices) -> FunctionType:
        '''
            GaussianLikeNoiseGenerator method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'   

            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''

        @qml.qnode(dev)
        def qnode() -> dict:
            '''
                PennyLane based quantum circuit composed of a Ry gate

                Parameters:  
                -----------  
                Nothing

                Returns:  
                --------  
                - counts : dict  
                    Sample from the supplied observable, with the number of shots determined from the dev.shots attribute of the corresponding device, 
                    returning the number of counts for each sample. If no observable is provided then basis state samples are returned directly 
                    from the device. Note that the output shape of this measurement process depends on the shots specified on the device. 
            '''

            qml.RandomLayers(np.random.random((len(dev.wires),self.n_layers)), range(len(dev.wires)))
            return qml.expval(qml.PauliZ(0))

        return qnode()

class RandomCZNoiseGenerator(QuantumCircuit):
    '''
        This class implements a quantum noise generator using a random circuit
    '''
    
    def __init__(self, location : float, scale : float, n_qubits: int, n_layers : int, n_entangling_layers : int = 2, bit_string : bool = True, state_id : int = 0, dev : qml.devices = None) -> None:
        '''
            RandomCircuitNoiseGenerator constructor.  

            Parameters:  
            -----------
            - location : float  
                location parameter translate the pdf, relative to the standard normal distribution
            - scale : float  
                scale parameter that stretches the pdf. The greater the magnitude, the greater the stretching. 
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit 
            - n_entangling_layers : int  
                number of layers with entanglement
            - state_id : int
                name of the state to consider
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'. Recommendent qml.device("default.mixed", wires=1, shots=1000).
            
            Returns:  
            --------  
            Nothing, a RandomCircuitNoiseGenerator object will be created.  
        '''

        super().__init__(n_qubits=n_qubits, n_layers=n_layers, dev=dev)

        self.shots               = dev.shots
        self.location            = location 
        self.scale               = scale
        self.n_qubits            = n_qubits
        self.n_layers            = n_layers
        self.n_entangling_layers = n_entangling_layers
        self.bit_string          = bit_string
        self.state_id            = state_id
        self.dev                 = dev
        self.circuit             = self.circ(self.dev)
        
    def generate_noise(self):
        '''
            GaussianLikeNoiseGenerator that generates on observable  

            Parameters:  
            -----------
            Nothing
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise observable
        '''
        if self.bit_string:
            unitary_probs = self.circuit()
            state_id = np.random.choice(2**self.n_qubits, p=unitary_probs)
            expval = unitary_probs[state_id]
        else:
            expval = self.circuit()[self.state_id]
        
        n = self.location + self.scale*expval
        return n

    
    def generate_noise_array(self, shape:tuple):
        '''
            GaussianLikeNoiseGenerator that generates a matrix of observables

            Parameters:  
            -----------
            shape : tuple
                shape of the desired maxtrix of observables
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise matrix of observables
        '''
        n = []
        for _ in range(np.prod(shape)): n.append(self.generate_noise())
        n = np.array(n)
    
        return n.reshape(shape)
    
    def circ(self, dev : qml.devices) -> FunctionType:
        '''
            GaussianLikeNoiseGenerator method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'   

            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''

        @qml.qnode(dev)
        def qnode() -> dict:
            '''
                PennyLane based quantum circuit composed of a Ry gate

                Parameters:  
                -----------  
                Nothing

                Returns:  
                --------  
                - counts : dict  
                    Sample from the supplied observable, with the number of shots determined from the dev.shots attribute of the corresponding device, 
                    returning the number of counts for each sample. If no observable is provided then basis state samples are returned directly 
                    from the device. Note that the output shape of this measurement process depends on the shots specified on the device. 
            '''

            for _ in range(self.n_layers):
                for j in range(self.n_qubits):
                    
                    qml.Hadamard(wires=j)
                    rZ = 2*np.pi*random.random()
                    qml.RZ(rZ,wires=j)

                    rY = 2*np.pi*random.random()
                    qml.RY(rY,wires=j)

                    for _ in range(self.n_entangling_layers):
                        r = random.sample(range(0, self.n_qubits), 2)        
                        qml.CZ(wires=r)

                    #qml.CZ(wires=r)
                # Hadamards
                qml.Barrier()
            for i in range(self.n_qubits): qml.Hadamard(wires=i)
        
            return qml.probs(wires=range(0,self.n_qubits))

        return qnode


import torch
import sys

sys.path += ['.', './utils/', '/circuits/']

from hqm.circuits.circuit import QuantumCircuit
from hqm.utils.aiinterface import AIInterface


class QGRU(torch.nn.Module):
    '''
        Quantum Gradient Recurrent Unit layer.

        Currently supports only Torch.

        Reference
        ---------
        A. Ceschini, A. Rosato and M. Panella, "Hybrid Quantum-Classical Recurrent  
        Neural Networks for Time Series Prediction," 2022 International Joint Conference  
        on Neural Networks (IJCNN), Padua, Italy, 2022, pp. 1-8,   
        doi: 10.1109/IJCNN55064.2022.9892441.
    '''

    def __init__(self, qcircuits : list[QuantumCircuit], inputsize : int, hiddensize : int, aiframework : str = 'torch') -> None:
        '''
        QGRU constructor.  

        Parameters:  
        -----------   
        - qcircuits : list of QuantumCircuit  
            list containing three quantum circuits in this exact order: 1) Quantum Layer Reset, 2) Quantum Layer Update, 3) Quantum Layer Output
        - inputsize : int  
            integer representing the number of variable (channels) in the input date
        - hiddensize : int  
            integer size representing the recurrent filters
        - aiframework : str  
            string representing the AI framework in use, can be 'torch' or 'keras'. This will create  
            a compatible trainable layer for the framework.

        Returns:    
        --------     
        Nothing, a QGRU object will be created.  
        '''
        
        super().__init__()

        if aiframework not in ['torch', 'keras']: raise Exception(f"Quanvolution2D curently supports only 'torch' as framework, found {aiframework}")
        if inputsize       < 1:                   raise Exception(f"inputsize must be greater than 1, found {inputsize}")
        if hiddensize      < 1:                   raise Exception(f"hiddensize must be greater than 1, found {hiddensize}")
        if len(qcircuits) != 3:                   raise Exception(f"qcircuits must contain 3 elements, one for reset gate, one for update gate and one for output gate, found {len(qcircuits)}")

        if (qcircuits[0].n_qubits != qcircuits[1].n_qubits) or (qcircuits[0].n_qubits != qcircuits[2].n_qubits) or (qcircuits[1].n_qubits != qcircuits[2].n_qubits) or (qcircuits[1].n_qubits != qcircuits[2].n_qubits):
            raise Exception(f"n_qubits must be the same for each circuit in qcircuits, found {qcircuits[0].n_qubits}, {qcircuits[1].n_qubits} and {qcircuits[2].n_qubits}")

        self.aiframework    = aiframework
        self.n_qubits       = qcircuits[0].n_qubits
        self.hiddensize     = hiddensize

        
        self.clayer_in      = torch.nn.Linear(inputsize+hiddensize, self.n_qubits)
        self.clayer_out     = torch.nn.Linear(self.n_qubits, hiddensize)

        self.qlayer_reset   = AIInterface.network_layer(
                                circuit      = qcircuits[0].circuit, 
                                weight_shape = qcircuits[0].weight_shape, 
                                n_qubits     = self.n_qubits, 
                                aiframework  = self.aiframework
                            )

        self.qlayer_update  = AIInterface.network_layer(
                                circuit      = qcircuits[1].circuit, 
                                weight_shape = qcircuits[1].weight_shape, 
                                n_qubits     = self.n_qubits, 
                                aiframework  = self.aiframework
                            )
        
        self.qlayer_output  = AIInterface.network_layer(
                                circuit      = qcircuits[2].circuit, 
                                weight_shape = qcircuits[2].weight_shape, 
                                n_qubits     = self.n_qubits, 
                                aiframework  = self.aiframework
                            )
        
        
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
        Torch forward function for QGRU layer

        Parameters:
        -----------
        - x : torch.Tensor
            input image or tensor
        
        Returns:
        --------
        - out : torch.Tensor
            qgru input
        '''

        if len(x.shape) != 3: raise Exception(f"x must be a tensor of 3 elements (batch, sequencelenght, featuressize), found {len(x.shape)}")

        batch_size, seq_length, featuressize = x.size()            
        hidden_seq = []
        h_t = torch.zeros(batch_size, self.hiddensize)
        
        for t in range(seq_length):
             # get features from the t-th element in seq, for all entries in the batch
            x_t = x[:, t, :]

            # Concatenate input and hidden state
            v_t = torch.cat((h_t, x_t), dim=1)

            # match qubit dimension
            y_t = self.clayer_in(v_t)

            r_t = torch.sigmoid(self.clayer_out(self.qlayer_reset(y_t)))  # forget block
            z_t = torch.sigmoid(self.clayer_out(self.qlayer_update(y_t)))  # update block

            # Concatenate input and hidden state
            v2_t = torch.cat(((r_t * h_t), x_t), dim=1)

            # match qubit dimension
            y2_t = self.clayer_in(v2_t)
            
            h_tilde_t = torch.tanh(self.clayer_out(self.qlayer_output(y2_t)))

            h_t = ((1-z_t) * h_tilde_t) + (z_t * h_t)

            hidden_seq.append(h_t.unsqueeze(0))

        hidden_seq = torch.cat(hidden_seq, dim=0)
        hidden_seq = hidden_seq.transpose(0, 1).contiguous()

        out = hidden_seq
        return out


import sys
sys.path += ['.', './hqm/']

from hqm.circuits.angleencoding import BasicEntangledCircuit
from hqm.classification.hcnn import HybridLeNet5
from hqm.layers.basiclayer import BasicLayer

import pennylane as qml
import torch


def test_qtorch_cnn_classifier_output():

    N_QUBITS      = 12
    N_LAYERS      = 1
    BATCH_SIZE    = 16

    IN_SHAPE      = (3, 64,64)
    OU_DIM        = 10
    
    
    dev = qml.device("default.qubit", wires=N_QUBITS)
    qcircuit = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    model = HybridLeNet5(qlayer=BasicLayer(qcircuit, aiframework='torch'), in_shape=IN_SHAPE, ou_dim=OU_DIM)
    
    x = torch.rand((BATCH_SIZE,) + IN_SHAPE)
    o = model(x)

    assert o.shape == (BATCH_SIZE, OU_DIM)


import sys
sys.path += ['.', './layers/', '/circuits/', './utils/']

from hqm.layers.quanvolution import Quanvolution2D
from hqm.circuits.circuit import QuantumCircuit
from hqm.layers.basiclayer import BasicLayer
from hqm.utils.sizes import size_conv_layer

import torch


class QuanvolutionAutoencoder(torch.nn.Module): 
    '''
        Hybrdid convolutional autoencoder, the encoder is composed of quanvolution layers, the decoder is composed of classical layers.
    '''

    def __init__(self, qcircuits : list[QuantumCircuit], in_shape : tuple[int], filters : list[int], kernelsizes : list[int], strides : list[int]) -> None:
        '''
            QuanvolutionAutoencoder constructor.  

            Parameters:  
            -----------
            - qcircuits : list  
                list of QuantumCircuit objects
            - in_shape : tuple 
                tuple represeting the image shape, (width, height, channels)
            - filters : list
                list containing the number of filters for each quanvolution layer
            - kernelsizes: list
                list containing the kernelsize for each quanvolution layer
            - strides: list
                list containin the stride for each quanvolution layer  
            
            Returns:  
            --------  
            Nothing, a QuanvolutionAutoencoder object will be created. 
        '''

        super().__init__()

        if len(filters)     != len(qcircuits):   raise Exception(f"lenght of filters must be the same of lenght of qcircuits, found {len(filters)} and {len(qcircuits)}")
        if len(kernelsizes) != len(qcircuits):   raise Exception(f"lenght of kernelsizes must be the same of lenght of qcircuits, found {len(kernelsizes)} and {len(qcircuits)}")
        if len(strides)     != len(qcircuits):   raise Exception(f"lenght of strides must be the same of lenght of qcircuits, found {len(strides)} and {len(qcircuits)}")
        if len(filters)     != len(kernelsizes): raise Exception(f"lenght of filters must be the same of lenght of kernelsizes, found {len(filters)} and {len(kernelsizes)}")
        if len(filters)     != len(strides):     raise Exception(f"lenght of filters must be the same of lenght of strides, found {len(filters)} and {len(strides)}")
        if len(strides)     != len(kernelsizes): raise Exception(f"lenght of strides must be the same of lenght of kernelsizes, found {len(strides)} and {len(kernelsizes)}")
        if len(in_shape)    != 3:                raise Exception(f"length of in_shape must be equals to 3 (widht, height, channels, found {len(in_shape)}")
        
        self.encoder  = []
        self.decoder  = []
        self.sizes    = []
        
        w, h, c       = in_shape
        self.depth    = len(qcircuits)
        
        # Building the quanvolution encoder
        for i in range(self.depth):
            self.encoder.append(Quanvolution2D(qcircuit=qcircuits[i], filters=filters[i], kernelsize=kernelsizes[i], stride=strides[i]))
        
        filters.reverse()
        filters.append(filters[-1])
        kernelsizes.reverse()
        strides.reverse()

        # Building the classical decoder
        for i in range(self.depth):
            self.decoder.append(torch.nn.ConvTranspose2d(filters[i], filters[i+1], kernel_size=kernelsizes[i], stride=strides[i]))
        

    def encoder_f(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method for the encoder

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''
        
        for i in range(self.depth):
            x = torch.nn.functional.relu(self.encoder[i](x))
        
        out = x
        return out

    def decoder_f(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method for the decoder

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        for i in range(self.depth - 1):
            x = torch.nn.functional.relu(self.decoder[i](x))
        out = torch.nn.functional.sigmoid(self.decoder[-1](x))
        
        return out

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        dec = self.encoder_f(x)
        out = self.decoder_f(dec)
        
        return out
        
class HybridAutoencoder(torch.nn.Module): 
    '''
        Hybrdid convolutional autoencoder, the encoder and the decodere are composed of clasical layer, the hidenn space
        is processed by a quantum circuit.
    '''

    def __init__(self, qlayer : BasicLayer, in_shape : tuple[int], filters : list[int], kernelsizes : list[int], strides : list[int]) -> None:
        '''
            QuanvolutionAutoencoder constructor.  

            Parameters:  
            -----------
            - qlayer : hqm.layers.basilayer.BasicLayer  
                hqm quantum layer to be stacked between two fully connected layers
            - in_shape : tuple 
                tuple represeting the image shape, (width, height, channels)
            - filters : list
                list containing the number of filters for each quanvolution layer
            - kernelsizes: list
                list containing the kernelsize for each quanvolution layer
            - strides: list
                list containin the stride for each quanvolution layer  
            
            Returns:  
            --------  
            Nothing, a QuanvolutionAutoencoder object will be created. 
        '''

        super().__init__()

        if len(filters)     != len(kernelsizes): raise Exception(f"lenght of filters must be the same of lenght of kernelsizes, found {len(filters)} and {len(kernelsizes)}")
        if len(filters)     != len(strides):     raise Exception(f"lenght of filters must be the same of lenght of strides, found {len(filters)} and {len(strides)}")
        if len(strides)     != len(kernelsizes): raise Exception(f"lenght of strides must be the same of lenght of kernelsizes, found {len(strides)} and {len(kernelsizes)}")
        if len(in_shape)    != 3:                raise Exception(f"length of in_shape must be equals to 3 (widht, height, channels, found {len(in_shape)}")
        
        self.encoder  = []
        self.decoder  = []
        self.sizes    = []
        
        w, h, c       = in_shape
        self.depth    = len(filters)

        self.encoder.append(torch.nn.Conv2d(c, filters[0], kernel_size=kernelsizes[0], stride=strides[0]))
        w = size_conv_layer(s=w, kernel_size=kernelsizes[0], padding=0, stride=strides[0])
        h = size_conv_layer(s=h, kernel_size=kernelsizes[0], padding=0, stride=strides[0])
        c = filters[0]
        # Building the quanvolution encoder
        for i in range(1, self.depth):
            self.encoder.append(torch.nn.Conv2d(filters[i-1], filters[i], kernel_size=kernelsizes[i], stride=strides[i]))
            w = size_conv_layer(s=w, kernel_size=kernelsizes[i], padding=0, stride=strides[i])
            h = size_conv_layer(s=h, kernel_size=kernelsizes[i], padding=0, stride=strides[i])
            c = filters[i]
        
        self.w            = w
        self.h            = h
        self.c            = c
        self.flatten_size = w*h*c
        self.fc1          = torch.nn.Linear(self.flatten_size, qlayer.n_qubits)
        self.qc_1         = qlayer.qlayer
        self.fc2          = torch.nn.Linear(qlayer.n_qubits, self.flatten_size)
        
        filters.reverse()
        filters.append(filters[-1])
        kernelsizes.reverse()
        strides.reverse()

        # Building the classical decoder
        for i in range(self.depth):
            self.decoder.append(torch.nn.ConvTranspose2d(filters[i], filters[i+1], kernel_size=kernelsizes[i], stride=strides[i]))
        

    def encoder_f(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method for the encoder

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        for i in range(self.depth):
            x = torch.nn.functional.relu(self.encoder[i](x))

        x = x.view(-1, self.flatten_size)
        x = torch.nn.functional.relu(self.fc1(x))
        x = torch.nn.functional.relu(self.qc_1(x))
        x = torch.nn.functional.relu(self.fc2(x))
        out = x.reshape((x.shape[0], self.c, self.w, self.h))
        return out

    def decoder_f(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method for the decoder

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        for i in range(self.depth - 1):
            x = torch.nn.functional.relu(self.decoder[i](x))
        out = torch.nn.functional.sigmoid(self.decoder[-1](x))
        
        return out

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        dec = self.encoder_f(x)
        out = self.decoder_f(dec)
        
        return out
        


import sys
sys.path += ['.', './hqm/']

from hqm.circuits.angleencoding import BasicEntangledCircuit, StronglyEntangledCircuit, RandomCircuit

import numpy as np

def test_BasicEntangledCircuit_output():
    n_qubits = 2
    n_layers = 1
    circ = BasicEntangledCircuit(n_qubits=n_qubits, n_layers=n_layers).circuit
    cinput   = np.zeros((n_qubits))
    cweights = np.zeros((n_layers,n_qubits))
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits


def test_StronglyEntangledCircuit_output():
    n_qubits = 2
    n_layers = 1
    circ = StronglyEntangledCircuit(n_qubits=n_qubits, n_layers=n_layers).circuit
    cinput   = np.zeros((n_qubits))
    cweights = np.zeros((n_layers,n_qubits, 3))
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits


def test_RandomCircuit_output():
    n_qubits = 2
    n_layers = 1
    circ = RandomCircuit(n_qubits=n_qubits, n_layers=n_layers).circuit
    cinput   = np.zeros((n_qubits))
    cweights = np.zeros((n_layers,n_qubits))
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits


import sys
sys.path += ['.', './layers/']

from hqm.layers.basiclayer import BasicLayer
import torch

class BasicHybridMLPClassifier(torch.nn.Module):
    '''
        This class implements a basic hybrid multilayer perceptron for classification purposes.
        BasicHybridMLPClassifier is composed of quantum layers stacked between two fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayer : BasicLayer, in_dim : int, ou_dim : int) -> None:
        '''
            BasicHybridMLPClassifier constructor.  

            Parameters:  
            -----------  
            - qlayer : hqm.layers.basiclayers.BasicLayer  
                hqm quantum layer to be stacked between two fully connected layers  
            - in_dim : int  
                integer representing the input size for the first fully connected layer  
            - ou_dim : int   
                integer representing the output size of the hybrid model  
            
            Returns:  
            --------  
            Nothing, a BasicHybridMLPClassifier object will be created.    
        '''
        super().__init__()

        if in_dim < 1: raise Exception(f"The parameter in_dim must be greater than 1, found {in_dim}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")

        n_qubits     = qlayer.n_qubits
        self.fc_1    = torch.nn.Linear(in_dim, n_qubits)
        self.qc_1    = qlayer.qlayer
        self.fc_2    = torch.nn.Linear(n_qubits, ou_dim)
        self.tanh    = torch.nn.Tanh()
        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''
        x = self.fc_1(x)
        x = self.tanh(x)
        x = self.qc_1(x)
        x = self.tanh(x)
        x = self.fc_2(x)
        out = self.softmax(x)
        return out
    
class MultiHybridMLPClassifier(torch.nn.Module):
    '''
        This class implements a hybrid multilayer perceptron with multiple quantum circuits for classification purposes.
        MultiHybridMLPClassifier is composed of several quantum layers stacked between two fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayers : list, in_dim : int, ou_dim : int) -> None:
        '''
            MultiHybridMLPClassifier constructor.  

            Parameters:  
            -----------  
            - qlayers : list  
                list of hqm quantum layers to be stacked between two fully connected layers  
            - in_dim : int  
                integer representing the input size for the first fully connected layer  
            - ou_dim : int  
                integer representing the output size of the hybrid model  
            
            Returns:  
            --------  
            Nothing, a MultiHybridMLPClassifier object will be created.    
        '''

        super().__init__()

        if in_dim < 1: raise Exception(f"The parameter in_dim must be greater than 1, found {in_dim}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")
        if len(qlayers) < 1: raise Exception(f"Size of qlayers must be greater than 1, found {len(qlayers)}")

        n_qubits_0   = qlayers[0].n_qubits
        n_qubits_1   = qlayers[-1].n_qubits
        self.fc_1    = torch.nn.Linear(in_dim, n_qubits_0)
        self.qcs     = [circ.qlayer for circ in qlayers]
        self.fc_2    = torch.nn.Linear(n_qubits_1, ou_dim)
        self.tanh    = torch.nn.Tanh()
        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor   
                output from the torch model  
        '''

        x = self.fc_1(x)
        x = self.tanh(x)
        for qc in self.qcs: 
            x = qc(x)
            x = self.tanh(x)
        x = self.fc_2(x)
        out = self.softmax(x)
        return out

class MultiHybridMultiMLPClassifier(torch.nn.Module):
    '''
        This class implements a hybrid multilayer perceptron with multiple quantum circuits for classification purposes.
        MultiHybridMultiMLPClassifier is composed of several quantum layers stacked between alternating fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayers : list, in_dims : list, ou_dim : list) -> None:
        '''
            MultiHybridMultiMLPClassifier constructor.  

            Parameters:  
            -----------  
            - qlayers : list  
                list of hqm quantum layers to be stacked between two fully connected layers  
            - in_dims: list  
                list of integers representing the input size for the i-th fully connected layer (first value should correspond to size of input data)  
            - ou_dim : list  
                list of integers representing the output size for the i-th fully connected layer (last value should correspond to desired output size)  
            
            Returns:  
            --------  
            Nothing, a MultiHybridMultiMLPClassifier object will be created.    
        '''
        
        super().__init__()

        if len(in_dims) < 1: raise Exception(f"Size in_dims must be greater than 1, found {len(in_dims)}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")
        if len(qlayers) < 1: raise Exception(f"Size of qlayerss must be greater than 1, found {len(qlayers)}")
        if len(qlayers) != len(in_dims): raise Exception(f"qlayers and in_dims must have the same lenght, found {len(qlayers)} and {len(in_dims)}")
        for i, dim in enumerate(in_dims): 
            if dim < 1: raise Exception(f"Element {i} of in_dims must be greater than 1, found {dim}")
            else: pass
        
        self.fcs     = [torch.nn.Linear(dim, circ.n_qubits) for (dim, circ) in zip(in_dims, qlayers)]
        self.fco     = torch.nn.Linear(qlayers[-1].n_qubits, ou_dim)
        self.qcs     = [circ.qlayer for circ in qlayers]
        self.tanh    = torch.nn.Tanh()
        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor   
                input for the torch model  
 
            Returns:  
            --------  
            - x : torch.Tensor  
                output from the torch model  
        '''
        
        for fc, qc in zip(self.fcs, self.qcs):
            x = fc(x)
            x = self.tanh(x)
            x = qc(x)
            x = self.tanh(x)
        
        x   = self.fco(x) 
        out = self.softmax(x)
        return out


import torchvision
import torch
import sys

sys.path += ['.', './utils/', '/circuits/']

from hqm.circuits.circuit import QuantumCircuit
from hqm.utils.aiinterface import AIInterface


class Quanvolution2D(torch.nn.Module):
    '''
        Quanvolution2D layer.

        Currently supports only Torch.
    '''

    def __init__(self, qcircuit : QuantumCircuit, filters : int, kernelsize : int = 3, stride : int = 1, padding : str = 'same', aiframework : str = 'torch') -> None:
        '''
        Quanvolution2D constructor.  

        Parameters:  
        -----------  
        - qcircuit : hqm.circuits.circuit.QuantumCircuit  
            QuantumCircuit object to be embedded into the quantum layer
        - filters : int
            number of quanvolution filters
        - kernelsize : int
            size of quanvolution kernel
        - stride : int
            stride for quanvolution operation
        - padding : str
            padding mode, same of valid
        - aiframework : str    
            string representing the AI framework in use, can be 'torch' or 'keras'. This will create  
            a compatible trainable layer for the framework.

        Returns:    
        --------     
        Nothing, a Quanvolution2D object will be created.  
        '''

        super().__init__()

        if aiframework not in ['torch', 'keras']: raise Exception(f"Quanvolution2D curently supports only 'torch' as framework, found {aiframework}")
        if kernelsize < 1:                        raise Exception(f"kernelsize must be greater than 1, found {kernelsize}")
        if stride < 1:                            raise Exception(f"stride must be greater than 1, found {stride}")
        
        self.aiframework = aiframework
        self.n_qubits    = qcircuit.n_qubits
        
        if kernelsize**2 > self.n_qubits:         raise Exception(f"kernelsize**2 must be lower than n_qubits, found kernelsize**2={kernelsize**2} and {self.n_qubits}")
        if filters > self.n_qubits:               raise Exception(f"filters must be lower than n_qubits, found {filters} and {self.n_qubits}")
        
        self.filters    = filters 
        self.kernelsize = kernelsize
        self.stride     = stride
        self.padding    = padding
        self.qlayer     = AIInterface.network_layer(
                                circuit      = qcircuit.circuit, 
                                weight_shape = qcircuit.weight_shape, 
                                n_qubits     = qcircuit.n_qubits, 
                                aiframework  = self.aiframework
                            )
        
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
        Torch forward function for quanvolution layer

        Parameters:
        -----------
        - x : torch.Tensor
            input image or tensor
        
        Returns:
        --------
        - out : torch.Tensor
            quanvoluted input
        '''

        if len(x.shape) != 4: raise Exception(f"x must be a tensor of 4 elements (batch, channels, width, height), found {len(x.shape)}")

        # Calculates the image shape after the convolution
        bs, ch, h, w = x.shape
        
        h_out = int(((h-self.kernelsize) / self.stride) +1)
        w_out = int(((w-self.kernelsize) / self.stride) +1)
        
        if self.padding == 'same':
            w_pad = int(round((w - w_out)/2))
            h_pad = int(round((h - h_out)/2))
        
        if self.padding == 'valid':
            h_pad = 0
            w_pad = 0 

        out = torch.zeros((bs, self.filters, h_out, w_out, ch))

        # Batch Loop
        for b in range(bs):
            # Channel Loop
            for c in range(ch):
                # Spatial Loops                                                
                for j in range(0, h_out, self.stride):
                    for k in range(0, w_out, self.stride):            
                        # Process a kernel_size*kernel_size region of the images
                        # with the quantum circuit stride*stride
                        p = x[b, c, j:j+self.kernelsize, k:k+self.kernelsize].reshape(-1)
                        q_results = self.qlayer(p)

                        for f in range(self.filters):
                            #out[b, f, j // self.kernelsize, k // self.kernelsize, c] = q_results[f]
                            out[b, f, j:j+self.kernelsize, k:k+self.kernelsize, c] = q_results[f]

        out = torch.mean(out, axis=-1)
        
        
        if self.padding == 'same':
            out = torch.nn.functional.pad(out, (h_pad, h_pad, w_pad, w_pad), "constant", 0)
            out = torchvision.transforms.Resize([w,h])(out)

        return out


import sys
sys.path += ['.', './hqm/']

from hqm.noise.gaussianlike import GaussianLikeNoiseGenerator
from hqm.noise.randomcircuit import RandomCircuitNoiseGenerator, RandomCZNoiseGenerator
import pennylane as qml

def test_GaussianLikeNoiseGenerator_output():
    gen = GaussianLikeNoiseGenerator(location=0, scale=1, dev=qml.device("default.mixed", wires=1, shots=1000))
    n = gen.generate_noise_array((64,30))
    assert n.shape == (64,30)

def test_RandomCircuitNoiseGenerator_output():
    N_QUBTS  = 2
    N_LAYERS = 20
    gen = RandomCircuitNoiseGenerator(location=0, scale=1, n_qubits=N_QUBTS, n_layers=N_LAYERS, dev=qml.device("default.qubit", wires=N_QUBTS))
    n = gen.generate_noise_array((64,30))
    assert n.shape == (64,30)

def test_RandomCZtNoiseGenerator_output():
    N_QUBTS  = 5
    N_LAYERS = 2
    gen = RandomCZNoiseGenerator(location=0, scale=1, n_qubits=N_QUBTS, n_layers=N_LAYERS, dev=qml.device("default.qubit", wires=N_QUBTS))
    n = gen.generate_noise_array((64,30))
    assert n.shape == (64,30)


if __name__ == '__main__':
    N_QUBTS  = 5
    N_LAYERS = 2
    gen = RandomCZNoiseGenerator(location=0, scale=1, n_qubits=N_QUBTS, n_layers=N_LAYERS, dev=qml.device("default.qubit", wires=N_QUBTS))
    n = gen.generate_noise_array((32,32))

    gen = RandomCZNoiseGenerator(location=0, scale=3, n_qubits=N_QUBTS, n_layers=N_LAYERS, dev=qml.device("default.qubit", wires=N_QUBTS))
    n2 = gen.generate_noise_array((32,32))

    import matplotlib.pyplot as plt

    plt.figure()
    plt.hist(n2.flatten(), 200, label='n2')
    plt.hist(n.flatten(), 200, label='n1')
    plt.xlim([0,1])
    plt.legend()
    plt.show()


import sys
sys.path += ['.', './hqm/']

from hqm.circuits.angleencoding import BasicEntangledCircuit
from hqm.layers.recurrent import QGRU
from hqm.utils.printer import Printer

import pennylane as qml
import numpy as np
import argparse
import torch


def test_qgru_output():
    N_QUBITS      = 9
    N_LAYERS      = 1
    INPUTSIZE     = 1
    HIDDENSIZE    = 10
    TIME_DIM      = 20
    BATCH_SIZE    = 16

    dev = qml.device("lightning.qubit", wires=N_QUBITS, shots = 1000)
    qcircuit1 = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    qcircuit2 = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    qcircuit3 = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    qcircuits = [qcircuit1, qcircuit2, qcircuit3]
    qgru      = QGRU(qcircuits=qcircuits, inputsize=INPUTSIZE, hiddensize=HIDDENSIZE)
    x = torch.rand(size=(BATCH_SIZE,TIME_DIM,INPUTSIZE))
    o =  qgru(x)

    assert o.shape == (BATCH_SIZE,TIME_DIM,HIDDENSIZE)